# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module containing OpenQASM to IonQ JSON conversion function

"""
from __future__ import annotations

import re
import warnings
from typing import TYPE_CHECKING, Any, Union

import openqasm3.ast

from qbraid.passes.qasm.compat import convert_qasm_pi_to_decimal
from qbraid.programs import load_program
from qbraid.programs.gate_model.ionq import IONQ_NATIVE_GATES, IonQProgram
from qbraid.programs.gate_model.qasm2 import OpenQasm2Program
from qbraid.programs.gate_model.qasm3 import OpenQasm3Program
from qbraid.transpiler.annotations import weight
from qbraid.transpiler.exceptions import ProgramConversionError

if TYPE_CHECKING:
    from qbraid.programs.typer import IonQDictType, QasmStringType

ONE_QUBIT_NON_PARAM = {
    "x": "x",
    "not": "not",
    "y": "y",
    "z": "z",
    "h": "h",
    "s": "s",
    "si": "si",
    "t": "t",
    "ti": "ti",
    "v": "v",
    "vi": "vi",
    "sdg": "si",
    "tdg": "ti",
    "sx": "v",
    "sxdg": "vi",
}

ONE_QUBIT_PARAM_ROT = {
    "p": "rz",
    "rx": "rx",
    "ry": "ry",
    "rz": "rz",
}

ONE_QUBIT_PARAM_PHASE = {
    "gpi": "gpi",
    "gpi2": "gpi2",
}


IONQ_ONE_QUBIT_GATE_MAP = {**ONE_QUBIT_NON_PARAM, **ONE_QUBIT_PARAM_ROT, **ONE_QUBIT_PARAM_PHASE}

TWO_QUBIT_NON_PARAM = {
    "cnot": "cnot",
    "cx": "cnot",
    "swap": "swap",
}

TWO_QUBIT_PARAM_ANGLE = {
    "zz": "zz",
    "rzz": "zz",
}

TWO_QUBIT_PARAM_ANGLE_PHASE = {
    "ms": "ms",
}

IONQ_TWO_QUBIT_GATE_MAP = {
    **TWO_QUBIT_NON_PARAM,
    **TWO_QUBIT_PARAM_ANGLE,
    **TWO_QUBIT_PARAM_ANGLE_PHASE,
}

IONQ_THREE_QUBIT_GATE_MAP = {
    "ccnot": "cnot",
    "ccx": "cnot",
    "toffoli": "cnot",
}

IONQ_THREE_QUBIT_GATE_ALIASES = {
    "ccnot": "ccnot",
    "ccx": "ccnot",
    "toffoli": "ccnot",
}


def extract_params(statement: openqasm3.ast.QuantumGate) -> list[str]:
    """Extracts the parameter(s) from a QuantumGate statement.

    Args:
        statement (openqasm3.ast.QuantumGate): QuantumGate statement to extract parameter(s) from.

    Returns:
        Union[str, list[str]]: Parameter(s) extracted from the QuantumGate statement.
    """
    try:
        params: str = re.findall(r"\((.+)\)", openqasm3.dumps(statement))[0]
    except IndexError:
        return []

    return [p.strip(" ") for p in params.split(",")]


def _parse_float_in_range(
    value: str, gate_name: str, param_name: str, bounds: tuple[float, float]
) -> float:

    min_val, max_val = bounds

    err_msg = (
        f"Invalid {param_name} value '{value}' for the '{gate_name}' gate. "
        f"{param_name.capitalize()} must be a float between {min_val} and {max_val}."
    )
    try:
        value = float(value)
    except ValueError as err:
        raise ValueError(err_msg) from err

    if not min_val <= value <= max_val:
        raise ValueError(err_msg)

    return value


def _parse_phase(phase: str, gate_name: str) -> float:
    return _parse_float_in_range(phase, gate_name, "phase", (-1, 1))


def _parse_angle(angle: str, gate_name: str) -> float:
    return _parse_float_in_range(angle, gate_name, "angle", (0, 0.25))


# pylint: disable-next=too-many-statements
def _parse_gates(program: Union[OpenQasm2Program, OpenQasm3Program]) -> list[dict[str, Any]]:
    program_qubits = program.module._qubit_registers.items()
    original_program: openqasm3.ast.Program = program.module.original_program

    gates: list[dict[str, Any]] = []

    contains_native = False
    non_zz_native_gates = set(IONQ_NATIVE_GATES) - {"zz"}

    # pylint: disable-next=too-many-nested-blocks
    for statement in original_program.statements:
        if isinstance(statement, openqasm3.ast.QuantumGate):
            name = statement.name.name.lower()

            if name == "id":
                continue

            contains_native = (
                contains_native
                or name in non_zz_native_gates
                or (name.startswith("c") and name[1:] in non_zz_native_gates)
            )

            qubits = statement.qubits
            qubit_values = []

            if len(qubits) == 1 and isinstance(qubits[0], openqasm3.ast.Identifier):
                reg_name = qubits[0].name
                for qreg_name, reg_size in program_qubits:
                    if qreg_name == reg_name:
                        qubit_values = list(range(reg_size))
                        break
            else:
                for qubit in qubits:
                    indices = qubit.indices
                    for index in indices:
                        qubit_values.extend(literal.value for literal in index)

            if name in IONQ_ONE_QUBIT_GATE_MAP:
                ionq_name = IONQ_ONE_QUBIT_GATE_MAP[name]
                if ionq_name in ONE_QUBIT_PARAM_ROT:
                    try:
                        angle: str = extract_params(statement)[0]
                    except IndexError as err:
                        raise ValueError(
                            f"Rotation parameter is required for the '{name}' "
                            "gate but was not provided."
                        ) from err
                    angle_decimal = float(convert_qasm_pi_to_decimal(angle))
                    for qubit in qubit_values:
                        gates.append(
                            {
                                "gate": ionq_name,
                                "target": qubit,
                                "rotation": angle_decimal,
                            }
                        )
                elif ionq_name in ONE_QUBIT_PARAM_PHASE:
                    try:
                        phase: str = extract_params(statement)[0]
                    except IndexError as err:
                        raise ValueError(
                            f"Phase parameter is required for the '{name}' "
                            "gate but was not provided."
                        ) from err
                    phase = _parse_phase(phase, ionq_name)

                    for qubit in qubit_values:
                        gates.append(
                            {
                                "gate": ionq_name,
                                "target": qubit,
                                "phase": phase,
                            }
                        )
                else:
                    for qubit in qubit_values:
                        gates.append({"gate": ionq_name, "target": qubit})

            elif name in IONQ_TWO_QUBIT_GATE_MAP:
                ionq_name = IONQ_TWO_QUBIT_GATE_MAP[name]

                if len(qubit_values) != 2:
                    raise ValueError(
                        f"Invalid number of qubits for the '{name}' gate. "
                        f"Expected 2, got {len(qubit_values)}"
                    )

                if ionq_name in TWO_QUBIT_PARAM_ANGLE:
                    try:
                        angle = extract_params(statement)[0]
                    except IndexError as err:
                        raise ValueError(
                            f"Angle parameter is required for the '{name}' "
                            "gate but was not provided."
                        ) from err

                    # Treat zz as 'qis' gate if all other gates are 'qis' gates
                    if name == "rzz" or (
                        ionq_name == "zz" and len(gates) > 0 and contains_native is False
                    ):
                        angle = float(convert_qasm_pi_to_decimal(angle))
                        gates.append(
                            {
                                "gate": ionq_name,
                                "rotation": angle,
                                "targets": qubit_values,
                            }
                        )

                    else:
                        key = "angle"

                        try:
                            angle = _parse_angle(angle, ionq_name)
                        except ValueError as err:
                            #  Treat zz with angle not in [0, 0.25] as 'qis'
                            if ionq_name == "zz" and contains_native is False:
                                key = "rotation"

                                try:
                                    angle = float(convert_qasm_pi_to_decimal(angle))
                                except ValueError:  # pylint: disable=raise-missing-from
                                    raise err
                            else:
                                raise err
                        else:
                            if ionq_name == "zz":
                                contains_native = True

                        gates.append(
                            {
                                "gate": ionq_name,
                                key: angle,
                                "targets": qubit_values,
                            }
                        )

                elif ionq_name in TWO_QUBIT_PARAM_ANGLE_PHASE:
                    params = extract_params(statement)
                    if len(params) not in {2, 3}:  # pragma: no cover
                        raise ValueError(
                            f"Invalid number of parameters for the '{name}' gate. "
                            f"Expected 2 or 3, got {len(params)}"
                        )

                    phases = [
                        _parse_phase(param, ionq_name) if i < 2 else _parse_phase(param, ionq_name)
                        for i, param in enumerate(params[:2])
                    ]
                    angle = _parse_angle(params[2], ionq_name) if len(params) == 3 else None

                    gate_data = {
                        "gate": ionq_name,
                        "phases": phases,
                        "targets": qubit_values,
                        **({"angle": angle} if angle is not None else {}),
                    }
                    gates.append(gate_data)

                elif ionq_name.startswith("c"):
                    gates.append(
                        {
                            "gate": ionq_name,
                            "control": qubit_values[0],
                            "target": qubit_values[1],
                        }
                    )
                else:
                    gates.append({"gate": ionq_name, "targets": qubit_values})

            elif name.startswith("c") and name[1:] in IONQ_ONE_QUBIT_GATE_MAP:
                ionq_name = IONQ_ONE_QUBIT_GATE_MAP[name[1:]]

                if len(qubit_values) != 2:
                    raise ValueError(
                        f"Invalid number of qubits for the '{name}' gate. "
                        f"Expected 2, got {len(qubit_values)}"
                    )

                if ionq_name in ONE_QUBIT_PARAM_ROT:
                    try:
                        angle: str = extract_params(statement)[0]
                    except IndexError as err:
                        raise ValueError(
                            f"Rotation parameter is required for the '{name}' "
                            "gate but was not provided."
                        ) from err
                    angle_decimal = float(convert_qasm_pi_to_decimal(angle))

                    gates.append(
                        {
                            "gate": ionq_name,
                            "control": qubit_values[0],
                            "target": qubit_values[1],
                            "rotation": angle_decimal,
                        }
                    )
                elif ionq_name in ONE_QUBIT_PARAM_PHASE:
                    try:
                        phase: str = extract_params(statement)[0]
                    except IndexError as err:
                        raise ValueError(
                            f"Phase parameter is required for the '{name}' "
                            "gate but was not provided."
                        ) from err
                    phase = _parse_phase(phase, ionq_name)

                    gates.append(
                        {
                            "gate": ionq_name,
                            "control": qubit_values[0],
                            "target": qubit_values[1],
                            "phase": phase,
                        }
                    )
                else:
                    gates.append(
                        {"gate": ionq_name, "control": qubit_values[0], "target": qubit_values[1]}
                    )

            elif name in IONQ_THREE_QUBIT_GATE_MAP:
                ionq_name = IONQ_THREE_QUBIT_GATE_MAP[name]

                if len(qubit_values) != 3:
                    raise ValueError(
                        f"Invalid number of qubits for the '{name}' gate. "
                        f"Expected 3, got {len(qubit_values)}"
                    )
                gates.append(
                    {
                        "gate": ionq_name,
                        "controls": qubit_values[:2],
                        "target": qubit_values[2],
                    }
                )

            else:
                raise ValueError(f"Gate '{name}' not supported by IonQ")

    return gates


@weight(1)
def openqasm3_to_ionq(qasm: Union[QasmStringType, openqasm3.ast.Program]) -> IonQDictType:
    """Returns an IonQ JSON format representation the input OpenQASM program.

    Args:
        qasm (str or openqasm3.ast.Program): OpenQASM program to convert to IonQDict type.

    Returns:
        dict: IonQ JSON format equivalent to input OpenQASM string.

    Raises:
        ProgramConversionError: For failure to parse gate data from OpenQASM string.
    """
    program: Union[OpenQasm2Program, OpenQasm3Program] = load_program(qasm)

    if program._module.has_measurements():
        warnings.warn(
            "Circuit contains measurement gates, which will be ignored "
            "during conversion to the IonQDictType",
            UserWarning,
        )

    gates = _parse_gates(program)

    if len(gates) == 0:
        raise ProgramConversionError("Failed to parse gate data from OpenQASM string.")

    gateset = IonQProgram.determine_gateset(gates)

    return {
        "qubits": program.num_qubits,
        "circuit": gates,
        "gateset": gateset.value,
        "format": "ionq.circuit.v0",
    }



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
OpenQASM 3 conversions

.. currentmodule:: qbraid.transpiler.conversions.qasm3

Functions
----------

.. autosummary::
   :toctree: ../stubs/

   qasm3_to_braket
   qasm3_to_openqasm3
   qasm3_to_qiskit
   qasm3_to_pyqir
   qasm3_to_ionq
   autoqasm_to_qasm3

"""
from .qasm3_extras import autoqasm_to_qasm3, qasm3_to_pyqir
from .qasm3_to_braket import qasm3_to_braket
from .qasm3_to_ionq import qasm3_to_ionq
from .qasm3_to_openqasm3 import qasm3_to_openqasm3
from .qasm3_to_qiskit import qasm3_to_qiskit

__all__ = [
    "qasm3_to_braket",
    "qasm3_to_openqasm3",
    "qasm3_to_qiskit",
    "qasm3_to_pyqir",
    "qasm3_to_ionq",
    "autoqasm_to_qasm3",
]



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unit tests for qbraid.programs.cirq.CirqCircuit

"""
from typing import Any
from unittest.mock import Mock

import cirq
import pytest
from cirq import CNOT, Circuit, GridQubit, H, LineQubit, Moment, NamedQubit, X, Y, Z

from qbraid.interface import circuits_allclose
from qbraid.programs.exceptions import ProgramTypeError
from qbraid.programs.gate_model.cirq import CirqCircuit


def test_contiguous_line_qubits():
    """Test convert to contiguous method for line qubits."""
    circuit = Circuit()
    circuit.append(X(LineQubit(0)))
    circuit.append(Y(LineQubit(2)))
    circuit.append(Z(LineQubit(4)))

    program = CirqCircuit(circuit)
    program.remove_idle_qubits()

    expected_qubits = [LineQubit(0), LineQubit(1), LineQubit(2)]
    assert set(program.qubits) == set(expected_qubits)
    assert program.num_clbits == 0


def test_remove_idle_grid_qubits_row():
    """Test convert to contiguous method for grid qubits."""
    circuit = Circuit()
    circuit.append(X(GridQubit(0, 0)))
    circuit.append(Y(GridQubit(0, 2)))
    circuit.append(Z(GridQubit(0, 4)))

    program = CirqCircuit(circuit)
    program.remove_idle_qubits()

    expected_qubits = [GridQubit(0, 0), GridQubit(0, 1), GridQubit(0, 2)]
    assert set(program.qubits) == set(expected_qubits)


def test_remove_idle_grid_qubits_col():
    """Test convert to contiguous method for grid qubits."""
    circuit = Circuit()
    circuit.append(X(GridQubit(0, 0)))
    circuit.append(Y(GridQubit(2, 0)))
    circuit.append(Z(GridQubit(4, 0)))

    program = CirqCircuit(circuit)
    program.remove_idle_qubits()

    expected_qubits = [GridQubit(0, 0), GridQubit(1, 0), GridQubit(2, 0)]
    assert set(program.qubits) == set(expected_qubits)


@pytest.mark.parametrize("index1,index2", [(1, 2), (2, 1)])
def test_remove_idle_qubits_raises(index1, index2):
    """Test removing idle qubits for grid qubits raises exception if
    row or col indicies are not all equal."""
    circuit = Circuit()
    circuit.append(X(GridQubit(0, 0)))
    circuit.append(Y(GridQubit(index1, index2)))

    program = CirqCircuit(circuit)

    with pytest.raises(ValueError):
        program.remove_idle_qubits()


def test_convert_grid_to_line_qubits():
    """Test converting grid qubits to line qubits."""
    circuit = Circuit()
    circuit.append(X(GridQubit(0, 0)))
    circuit.append(Y(GridQubit(0, 1)))
    circuit.append(Z(GridQubit(0, 2)))

    program = CirqCircuit(circuit)
    program._convert_to_line_qubits()

    expected_qubits = [LineQubit(0), LineQubit(1), LineQubit(2)]
    assert set(program.qubits) == set(expected_qubits)

    circuit = Circuit()
    circuit.append(X(GridQubit(0, 0)))
    circuit.append(Y(GridQubit(1, 0)))
    circuit.append(Z(GridQubit(2, 0)))

    program = CirqCircuit(circuit)
    program._convert_to_line_qubits()

    expected_qubits = [LineQubit(0), LineQubit(1), LineQubit(2)]
    assert set(program.qubits) == set(expected_qubits)

    circuit = Circuit()
    circuit.append(X(GridQubit(0, 0)))
    circuit.append(Y(GridQubit(1, 1)))
    circuit.append(Z(GridQubit(2, 2)))

    program = CirqCircuit(circuit)
    program._convert_to_line_qubits()

    expected_qubits = [LineQubit(0), LineQubit(1), LineQubit(2)]
    assert set(program.qubits) == set(expected_qubits)


def test_mixed_qubit_types():
    """Test convert to contiguous method raises error for mixed qubit type."""
    circuit = Circuit()
    circuit.append(X(LineQubit(3)))
    circuit.append(Y(GridQubit(0, 4)))

    program = CirqCircuit(circuit)
    program.remove_idle_qubits()

    expected_qubits = [LineQubit(0), LineQubit(1)]
    assert set(program.qubits) == set(expected_qubits)


def test_convert_named_qubit_to_line_qubit():
    """Test converting a circuit of all NamedQubits to LineQubits"""
    circuit = Circuit(
        [
            Moment(H(NamedQubit("q"))),
        ]
    )

    program = CirqCircuit(circuit)
    program._convert_to_line_qubits()
    for qubit in program.qubits:
        assert isinstance(qubit, LineQubit)


def test_reverse_named_qubit_order():
    """Test reversing the order of qubits in a NamedQubit circuit"""
    circuit = Circuit(
        [
            Moment(H(NamedQubit("q"))),
            Moment(X(NamedQubit("y"))),
        ]
    )

    op_map = {}
    for op in circuit.all_operations():
        op_map[str(op.gate)] = op.qubits

    program = CirqCircuit(circuit)
    program.reverse_qubit_order()
    rev_circuit = program.program
    for op in rev_circuit.all_operations():
        assert op_map[str(op.gate)] != op.qubits


def test_remove_idle_named_qubits():
    """Test removing idle qubits from a NamedQubit circuit"""
    circuit = Circuit(
        [
            Moment(H(NamedQubit("q_4")), H(NamedQubit("q_8"))),
            Moment(CNOT(NamedQubit("q_6"), NamedQubit("q_8"))),
            Moment(X(NamedQubit("q_2"))),
        ]
    )
    qprogram = CirqCircuit(circuit)
    qprogram.remove_idle_qubits()
    new_circuit = qprogram.program
    assert set(new_circuit.all_qubits()) == {NamedQubit(str(i)) for i in range(qprogram.num_qubits)}
    assert circuits_allclose(circuit, new_circuit)


def test_remove_all_measurements():
    """Test removing all measurement gates from a circuit"""
    q0, q1 = cirq.LineQubit.range(2)
    circuit = cirq.Circuit(
        cirq.H(q0), cirq.CNOT(q0, q1), cirq.measure(q0, key="m0"), cirq.measure(q1, key="m1")
    )
    expected_circuit = cirq.Circuit(cirq.H(q0), cirq.CNOT(q0, q1))
    result_circuit = CirqCircuit.remove_measurements(circuit)
    assert result_circuit == expected_circuit, "All measurement gates should be removed."


def test_no_measurements_to_remove():
    """Test removing measurements from a circuit with no measurements"""
    q0, q1 = cirq.LineQubit.range(2)
    circuit = cirq.Circuit(cirq.H(q0), cirq.CNOT(q0, q1))
    result_circuit = CirqCircuit.remove_measurements(circuit)
    assert (
        result_circuit == circuit
    ), "The circuit should remain unchanged as there are no measurements."


def test_mixed_operations_and_measurements():
    """Test removing measurements from a circuit with mixed operations and measurements"""
    q0, q1 = cirq.LineQubit.range(2)
    circuit = cirq.Circuit(
        cirq.H(q0), cirq.measure(q0, key="m0"), cirq.CNOT(q0, q1), cirq.measure(q1, key="m1")
    )
    expected_circuit = cirq.Circuit(cirq.H(q0), cirq.CNOT(q0, q1))
    result_circuit = CirqCircuit.remove_measurements(circuit)
    assert (
        result_circuit == expected_circuit
    ), "All measurement gates should be removed, leaving other operations intact."


def test_align_final_measurements():
    """Test aligning measurements for a circuit with measurements."""
    q0, q1 = cirq.LineQubit.range(2)
    circuit = cirq.Circuit(
        cirq.H(q0), cirq.CNOT(q0, q1), cirq.measure(q0, key="m0"), cirq.measure(q1, key="m1")
    )
    expected_circuit = cirq.Circuit(
        cirq.H(q0),
        cirq.CNOT(q0, q1),
        cirq.Moment(cirq.measure(q0, key="m0"), cirq.measure(q1, key="m1")),
    )
    aligned_circuit = CirqCircuit.align_final_measurements(circuit)
    assert (
        aligned_circuit == expected_circuit
    ), "The measurements should be aligned in the same moment"


def test_align_measurements_for_no_measurement():
    """Test aligning measurements for a circuit with no measurements."""
    q0, q1 = cirq.LineQubit.range(2)
    circuit = cirq.Circuit(cirq.H(q0), cirq.CNOT(q0, q1))
    expected_circuit = cirq.Circuit(cirq.H(q0), cirq.CNOT(q0, q1))
    aligned_circuit = CirqCircuit.align_final_measurements(circuit)
    assert (
        aligned_circuit == expected_circuit
    ), "The circuit should remain unchanged as there are no measurements"


def test_align_measurements_for_partial_measurement():
    """Test aligning measurements from a circuit where not all qubits are measured."""
    q0, q1 = cirq.LineQubit.range(2)
    circuit = cirq.Circuit(cirq.H(q0), cirq.CNOT(q0, q1), cirq.measure(q0, key="m0"))
    expected_circuit = circuit
    aligned_circuit = CirqCircuit.align_final_measurements(circuit)
    assert (
        aligned_circuit == expected_circuit
    ), "The circuit should remain unchanged as not all qubits are measured"


def test_raise_program_type_error():
    """Test raising ProgramTypeError"""
    with pytest.raises(ProgramTypeError):
        CirqCircuit("OPENQASM 2.0;qreg q[2];h q[0];cx q[0],q[1];")


def test_key_from_line_qubit():
    """Test generating a key from a LineQubit."""
    qubit = cirq.LineQubit(1)
    expected_key = "q(1)"
    assert CirqCircuit._key_from_qubit(qubit) == expected_key


def test_key_from_grid_qubit():
    """Test generating a key from a GridQubit."""
    qubit = cirq.GridQubit(3, 5)
    expected_key = "3"
    assert CirqCircuit._key_from_qubit(qubit) == expected_key


def test_key_from_named_qubit():
    """Test generating a key from a NamedQubit."""
    qubit = cirq.NamedQubit("qubit7")
    expected_key = "qubit7"
    assert CirqCircuit._key_from_qubit(qubit) == expected_key


def test_key_from_unsupported_qubit():
    """Test generating a key from an unsupported qubit type."""

    class UnsupportedQubit(cirq.Qid):
        """Unsupported qubit type for testing."""

        def _comparison_key(self) -> Any:
            """Dummy abstract method implemented for testing."""
            return 0

        def dimension(self) -> int:
            """Dummy abstract method implemented for testing."""
            return 0

    qubit = UnsupportedQubit()
    with pytest.raises(ValueError) as excinfo:
        CirqCircuit._key_from_qubit(qubit)

    expected_message = "Expected qubit of type 'GridQubit' 'LineQubit' or 'NamedQubit'"
    assert expected_message in str(excinfo.value)


def test_int_from_qubit_grid_qubit():
    """Test generating an integer key from a GridQubit."""
    qubit = cirq.GridQubit(3, 5)
    expected_key = 3
    assert CirqCircuit._int_from_qubit(qubit) == expected_key


def test_bad_qubit():
    """Test qubits that aren't a GridQubit, LineQubit, or NamedQubit."""
    qubit = "bad qubit"
    with pytest.raises(ValueError) as excinfo:
        CirqCircuit._int_from_qubit(qubit)
    expected_message = "Expected qubit of type 'GridQubit' 'LineQubit' or 'NamedQubit'"
    assert expected_message in str(excinfo.value)

    qubits = [qubit]
    targets = [0]
    with pytest.raises(ValueError) as excinfo:
        CirqCircuit._make_qubits(qubits, targets)


def test_transform_no_effect():
    """Test that transform method has no effect
    since it is not implemented"""
    qubit = cirq.LineQubit(0)
    circuit = cirq.Circuit(cirq.I(qubit))
    program = CirqCircuit(circuit)
    device = Mock()
    program.transform(device)
    assert program.program == circuit



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining input / output types for a quantum backend:

  * NATIVE_REGISTRY: Dict mapping all supported quantum software libraries / package
                     names to their respective program types.

"""

from copy import deepcopy
from importlib import import_module
from types import ModuleType
from typing import Any, Type

from .typer import BOUND_QBRAID_META_TYPES, QBRAID_META_TYPES


def _assign_default_type_alias(imported: dict[str, Any], program_type: Type[Any]) -> str:
    """
    Determines a unique alias for the given program type based on its module name.

    Args:
        imported (dict[str, Any]): A dictionary of already imported program type aliases.
        program_type (Type[Any]): The class or type for which to determine the alias.

    Returns:
        str: The determined alias for the program type.

    Raises:
        ValueError: If a unique alias cannot be determined due to duplicates.
    """
    module_name = program_type.__module__
    module_parts = module_name.split(".")
    alias = module_parts[0]

    if alias in imported:
        if len(module_parts) > 1:
            alias = f"{alias}_{module_parts[1]}"
        else:
            alias = f"{alias}_{program_type.__name__.lower()}"

        if alias in imported:
            raise ValueError(f"Duplicate alias {alias}")

    return alias


def _dynamic_importer(opt_modules: list[str]) -> dict[str, Type[Any]]:
    imported: dict[str, Type[Any]] = {}

    for m in opt_modules:
        try:
            data = m.split(".")
            for i, _ in enumerate(data):
                if data[:i]:
                    globals()[".".join(data[:i])] = import_module(".".join(data[:i]))
            # to be more secure, do not do module = globals()[m] = import_module(), as it could
            # create globals()[m] and later throw an error as there is no module named like that.
            # Is prefered to let module be import_module() and throw an exception if is needed
            module: ModuleType = import_module(m)
            globals()[m] = module
            program_type = _get_class(module.__name__)
            program_type_alias = _assign_default_type_alias(imported, program_type)
            program_type_alias = data[0] if program_type_alias == "builtins" else program_type_alias
            imported[program_type_alias] = program_type
        except Exception:  # pylint: disable=broad-except
            pass

    return imported


# pylint: disable=undefined-variable
def _get_class(module: str):
    if module == "bloqade.analog.builder.assign":
        return bloqade.analog.builder.assign.BatchAssign  # type: ignore # noqa: F821
    if module == "cirq":
        return cirq.Circuit  # type: ignore # noqa: F821
    if module == "qiskit":
        return qiskit.QuantumCircuit  # type: ignore # noqa: F821
    if module == "braket.ahs":
        return braket.ahs.AnalogHamiltonianSimulation  # type: ignore # noqa: F821
    if module == "braket.circuits":
        return braket.circuits.Circuit  # type: ignore # noqa: F821
    if module == "pennylane":
        return pennylane.tape.QuantumTape  # type: ignore # noqa: F821
    if module == "pyquil":
        return pyquil.Program  # type: ignore # noqa: F821
    if module == "pytket":
        return pytket._tket.circuit.Circuit  # type: ignore # noqa: F821
    if module == "openqasm3":
        return openqasm3.ast.Program  # type: ignore # noqa: F821
    if module == "pyqir":
        return pyqir.Module  # type: ignore # noqa: F821
    if module == "cpp_pyqubo":
        return cpp_pyqubo.Model  # type: ignore # noqa: F821
    if module == "cudaq":
        return cudaq.PyKernel  # type: ignore # noqa: F821
    if module == "qibo":  # pragma: no cover
        return qibo.Circuit  # type: ignore # noqa: F821
    if module == "stim":  # pragma: no cover
        return stim.Circuit  # type: ignore # noqa: F821
    if module == "pulser":
        return pulser.sequence.sequence.Sequence  # type: ignore # noqa: F821
    if module == "pyqpanda3":  # pragma: no cover
        return pyqpanda3.core.QProg  # type: ignore # noqa: F821
    if module == "autoqasm":
        return autoqasm.program.program.Program  # type: ignore # noqa: F821
    raise ValueError(f"Unsupported module '{module}'")


# pylint: enable=undefined-variable

# Supported quantum programs.
dynamic_type_registry: dict[str, Type[Any]] = _dynamic_importer(
    [
        "cirq",
        "qiskit",
        "pennylane",
        "pyquil",
        "pytket",
        "braket.circuits",
        "braket.ahs",
        "openqasm3",
        "cpp_pyqubo",
        "cudaq",
    ]
)
dynamic_non_native: dict[str, Type[Any]] = _dynamic_importer(
    ["bloqade.analog.builder.assign", "qibo", "stim", "pyqir", "pulser", "pyqpanda3", "autoqasm"]
)
static_type_registry: dict[str, Type[Any]] = {
    metatype.__alias__: metatype.__bound__ for metatype in BOUND_QBRAID_META_TYPES
}
qbraid_meta_type_registry: dict[str, Type[Any]] = {
    metatype.__alias__: metatype for metatype in QBRAID_META_TYPES
}

NATIVE_REGISTRY: dict[str, Type[Any]] = (
    dynamic_type_registry | static_type_registry | qbraid_meta_type_registry
)
_QPROGRAM_REGISTRY: dict[str, Type[Any]] = deepcopy(NATIVE_REGISTRY) | dynamic_non_native
_QPROGRAM_TYPES: set[Type[Any]] = set(_QPROGRAM_REGISTRY.values())
_QPROGRAM_ALIASES: set[str] = set(_QPROGRAM_REGISTRY.keys())



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module for interfacing with the Amazon Braket cost tracker

"""
from decimal import Decimal
from typing import Any, Optional

from boto3.session import Session
from braket.aws import AwsDevice, AwsQuantumTask, AwsSession
from braket.tracking.tracker import Tracker, _get_qpu_task_cost, _get_simulator_task_cost
from braket.tracking.tracking_context import register_tracker
from braket.tracking.tracking_events import _TaskCompletionEvent, _TaskCreationEvent


def _generate_creation_event(boto_data: dict[str, Any]) -> _TaskCreationEvent:
    """Generate a task creation event."""
    job_token = boto_data.pop("jobToken", None)
    return _TaskCreationEvent(
        arn=boto_data["quantumTaskArn"],
        shots=boto_data["shots"],
        is_job_task=(job_token is not None),
        device=boto_data["deviceArn"],
    )


def _generate_completion_event(
    boto_data: dict[str, Any], aws_session: Optional[AwsSession] = None
) -> _TaskCompletionEvent:
    """Generate a task completion event."""
    task_arn = boto_data["quantumTaskArn"]
    result = AwsQuantumTask(task_arn, aws_session=aws_session).result()
    execution_duration = None
    try:
        execution_duration = result.additional_metadata.simulatorMetadata.executionDuration
    except AttributeError:  # pragma: no cover
        pass

    completion_data = {
        "arn": task_arn,
        "status": boto_data["status"],
        "execution_duration": execution_duration,
    }
    return _TaskCompletionEvent(**completion_data)


def _get_tracker_task_details(
    boto_data: dict[str, Any], aws_session: Optional[AwsSession] = None
) -> dict[str, Any]:
    """Get the quantum task details populated by the Amazon Braket cost tracker."""
    tracker = Tracker()
    register_tracker(tracker)
    creation_event = _generate_creation_event(boto_data)
    completion_event = _generate_completion_event(boto_data, aws_session=aws_session)
    tracker.receive_event(creation_event)
    tracker.receive_event(completion_event)

    task_arn = boto_data["quantumTaskArn"]
    return tracker._resources[task_arn]


def get_quantum_task_cost(task_arn: str, aws_session: Optional[AwsSession] = None) -> Decimal:
    """Returns the cost of quantum task run on an AWS device

    Args:
        task_arn (str): The quantumTaskArn.
        aws_session (optional, braket.aws.AwsSession): Session with which to retrieve the task

    Returns:
        decimal.Decimal: The estimated total cost in USD

    Raises:
        ValueError: If quantum task not found or not in final state, or device type not supported.
    """
    task_region = task_arn.split(":")[3]

    if aws_session is None:
        boto_session = Session(region_name=task_region)
        aws_session = AwsSession(boto_session=boto_session)

    elif aws_session.region != task_region:
        raise ValueError(
            f"AwsSession region {aws_session.region} does not match task region {task_region}"
        )

    braket_client = aws_session.braket_client
    response = braket_client.get_quantum_task(quantumTaskArn=task_arn)
    status = response["status"]

    if status in ["FAILED", "CANCELLED"]:
        return Decimal(0)

    if status != "COMPLETED":
        raise ValueError(f"Task {task_arn} is not COMPLETED. Current state is {status}.")

    device_arn = response["deviceArn"]
    device = AwsDevice(arn=device_arn, aws_session=aws_session)
    device_type = device.type.value

    task_cost = {
        "SIMULATOR": _get_simulator_task_cost,
        "QPU": _get_qpu_task_cost,
    }

    details = _get_tracker_task_details(response, aws_session=aws_session)

    try:
        return task_cost[device_type](task_arn, details)
    except KeyError as err:
        raise ValueError(f"Device type {device_type} is not supported.") from err



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module containing sub-modules for interfacing with
various quantum software libraries and program types.

.. currentmodule:: qbraid.programs.ahs

Classes
--------

.. autosummary::
   :toctree: ../stubs/

   AnalogHamiltonianProgram
   AHSEncoder

Submodules
------------

.. autosummary::
   :toctree: ../stubs/

   braket_ahs

"""
import importlib

from ._model import AHSEncoder, AnalogHamiltonianProgram

_qbraid = importlib.import_module("qbraid.programs._import")
NATIVE_REGISTRY = getattr(_qbraid, "NATIVE_REGISTRY", {})

submodules = []
base_path = "qbraid.programs.ahs."

for lib in NATIVE_REGISTRY:
    try:
        imported_lib = importlib.import_module(base_path + lib)
        submodules.append(lib)
        globals()[lib] = imported_lib

    except ImportError:
        pass


__all__ = ["AnalogHamiltonianProgram", "AHSEncoder"]

__all__.extend(submodules)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining base program type specification

"""
from typing import Any, Callable, Optional, Type

from .experiment import ExperimentType
from .registry import (
    derive_program_type_alias,
    get_native_experiment_type,
    is_registered_alias_native,
    register_program_type,
)


class ProgramSpec:
    """Base class used to register program type and type alias."""

    def __init__(  # pylint: disable=too-many-arguments
        self,
        program_type: Type[Any],
        alias: Optional[str] = None,
        overwrite: bool = False,
        serialize: Optional[Callable[[Any], Any]] = None,
        validate: Optional[Callable[[Any], None]] = None,
        experiment_type: Optional[ExperimentType] = None,
    ):
        self._program_type = program_type
        self._serialize = serialize or (lambda program: program)
        self._validate = validate or (lambda program: None)

        register_program_type(program_type, alias=alias, overwrite=overwrite)
        self._alias = alias or derive_program_type_alias(program_type)
        self._native = is_registered_alias_native(self._alias)
        self._experiment_type: Optional[ExperimentType] = None
        self.experiment_type = experiment_type

    @property
    def program_type(self) -> Type[Any]:
        """Return the registered program type."""
        return self._program_type

    @property
    def alias(self) -> str:
        """Return the alias of the registered program type."""
        return self._alias

    @property
    def native(self) -> bool:
        """True if program is natively supported by qBraid, False otherwise."""
        return self._native

    @property
    def experiment_type(self) -> Optional[ExperimentType]:
        """Getter for experiment type."""
        return self._experiment_type

    @experiment_type.setter
    def experiment_type(self, value: Optional[ExperimentType]):
        """Setter for experiment type with logic for native aliases."""
        if value is not None:
            self._experiment_type = value
        elif self._native:
            self._experiment_type = get_native_experiment_type(self._alias)
        else:
            self._experiment_type = None

    def serialize(self, program: Any) -> Any:
        """
        Convert the given program to a format suitable for submission the
        qBraid API using the serialize lambda.

        Args:
            program (Any): The program to convert.

        Returns:
            Any: The serialized program, or the program itself if serialize is None.
        """
        return self._serialize(program)

    def validate(self, program: Any) -> None:
        """
        Validate the given program using the validate lambda.

        Args:
            program (Any): The program to validate.

        Raises:
            ValueError: If the program is invalid.
        """
        self._validate(program)

    def __str__(self) -> str:
        return f"ProgramSpec({self._program_type.__name__}, {self.alias})"

    def __repr__(self) -> str:
        return (
            f"<ProgramSpec('{self._program_type.__module__}.{self._program_type.__name__}', "
            f"'{self.alias}')>"
        )

    def __eq__(self, other: object) -> bool:
        """
        Compare this ProgramSpec object with another object for equality based on type and alias.

        Args:
            other (object): Another object to compare against.

        Returns:
            bool: True if both objects are instances of ProgramSpec and have the
                  same type and alias, False otherwise.
        """
        if not isinstance(other, ProgramSpec):
            return False

        this_spec = (self._program_type, self._alias, self._native)
        other_spec = (other._program_type, other._alias, other._native)
        return this_spec == other_spec



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module containing pytket programs used for testing

"""
from pytket.circuit import Circuit as TKCircuit


def pytket_bell() -> TKCircuit:
    """Returns pytket bell circuit"""
    circuit = TKCircuit(2)
    circuit.H(0)
    circuit.CX(0, 1)
    return circuit


def pytket_shared15():
    """Returns pytket `Circuit` for qBraid `TestSharedGates`."""

    circuit = TKCircuit(4)

    for i in range(4):
        circuit.H(i)
    for i in range(2):
        circuit.X(i)
    circuit.Y(2)
    circuit.Z(3)
    circuit.S(0)
    circuit.Sdg(1)
    circuit.T(2)
    circuit.Tdg(3)
    circuit.Rx(1 / 4, 0)
    circuit.Ry(1 / 2, 1)
    circuit.Rz(3 / 4, 2)
    circuit.U1(1 / 8, 3)
    circuit.SX(0)
    circuit.SXdg(1)
    circuit.ISWAPMax(2, 3)
    for i in range(2):
        circuit.SWAP(i, i + 2)
    circuit.CX(0, 1)
    circuit.CU1(1 / 4, 2, 3)

    return circuit



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

# pylint: disable=redefined-outer-name

"""
Unit tests for retrieving and post-processing experimental results.

"""
import datetime
from collections import Counter

import numpy as np
import pytest

from qbraid.programs import ExperimentType
from qbraid.runtime.native.result import NECVectorAnnealerResultData, QbraidQirSimulatorResultData
from qbraid.runtime.postprocess import (
    distribute_counts,
    format_data,
    normalize_batch_bit_lengths,
    normalize_bit_lengths,
    normalize_data,
)
from qbraid.runtime.result import Result
from qbraid.runtime.result_data import (
    AhsResultData,
    AhsShotResult,
    AnnealingResultData,
    GateModelResultData,
)


@pytest.fixture
def annealing_result_data() -> AnnealingResultData:
    """Fixture to create an AnnealingResultData object with some test data.
    Returns:
        AnnealingResultData: An AnnealingResultData object.
    """
    return AnnealingResultData(
        solutions=[
            {
                "spin": {" x1": 0, " x2": 0, "x1": 0},
                "energy": 0,
                "time": 0.006517000030726194,
                "constraint": True,
                "memory_usage": 1.189453125,
            }
        ]
    )


@pytest.fixture
def nec_vector_annealer_result_data() -> NECVectorAnnealerResultData:
    """Fixture to create an NECVectorAnnealerResultData object with some test data.
    Returns:
        NECVectorAnnealerResultData: An NECVectorAnnealerResultData object.
    """
    return NECVectorAnnealerResultData(
        [
            {
                "spin": {" x1": 0, " x2": 0, "x1": 0},
                "energy": 0,
                "time": 0.006517000030726194,
                "constraint": True,
                "memory_usage": 1.189453125,
            }
        ]
    )


@pytest.fixture
def gate_model_result_data():
    """Fixture to create a GateModelResultData object with some test data."""
    measurement_counts = {"00": 100, "01": 50, "10": 25}
    measurements = np.array([[0, 0], [0, 1], [1, 0]])
    return GateModelResultData(measurement_counts=measurement_counts, measurements=measurements)


@pytest.fixture
def shot_result() -> AhsShotResult:
    """Fixture to create an AhsShotResult object."""
    success = True
    pre_sequence = np.array([1, 1, 1, 1, 0, 1, 1, 1, 1, 1, 1])
    post_sequence = np.array([0, 0, 0, 1, 0, 1, 1, 0, 0, 1, 0])
    return AhsShotResult(success=success, pre_sequence=pre_sequence, post_sequence=post_sequence)


@pytest.fixture
def shot_result_failed() -> AhsShotResult:
    """Fixture to create an AhsShotResult object for failed shot."""
    success = False
    pre_sequence = None
    post_sequence = None
    return AhsShotResult(success=success, pre_sequence=pre_sequence, post_sequence=post_sequence)


@pytest.mark.parametrize(
    "success, pre_sequence, post_sequence, expected_dict",
    [
        (
            True,
            np.array([1, 1, 1, 1, 0, 1, 1, 1, 1, 1, 1]),
            np.array([0, 0, 0, 1, 0, 1, 1, 0, 0, 1, 0]),
            {
                "success": True,
                "pre_sequence": [1, 1, 1, 1, 0, 1, 1, 1, 1, 1, 1],
                "post_sequence": [0, 0, 0, 1, 0, 1, 1, 0, 0, 1, 0],
            },
        ),
        (
            False,
            None,
            None,
            {"success": False, "pre_sequence": None, "post_sequence": None},
        ),
    ],
)
def test_shot_result_to_dict(success, pre_sequence, post_sequence, expected_dict):
    """Test converting shot result data to dictionary format."""
    shot_result = AhsShotResult(
        success=success, pre_sequence=pre_sequence, post_sequence=post_sequence
    )
    assert shot_result.to_dict() == expected_dict


@pytest.fixture
def ahs_result_data(shot_result: AhsShotResult) -> AhsResultData:
    """Fixture to create an AhsResultData object."""
    measurements = [shot_result]
    state_counts = Counter()
    states = ["e", "r", "g"]

    for shot in measurements:
        if shot.success:
            pre = shot.pre_sequence
            post = shot.post_sequence
            state_idx = [
                0 if pre_i == 0 else 1 if post_i == 0 else 2 for pre_i, post_i in zip(pre, post)
            ]
            state = "".join(states[s_idx] for s_idx in state_idx)
            state_counts.update([state])

    measurement_counts = dict(state_counts)
    return AhsResultData(measurement_counts=measurement_counts, measurements=measurements)


@pytest.fixture
def qir_sim_data():
    """Fixture to create a QbraidQirSimulatorResultData object."""
    return QbraidQirSimulatorResultData(
        backend_version="0.7.4",
        seed=42,
        measurement_counts={"00": 10, "01": 15},
    )


@pytest.fixture
def result_instance():
    """Fixture to create a Result object for testing."""
    return Result(
        device_id="test_device",
        job_id="test_job",
        success=True,
        data=GateModelResultData(),
        test_detail="test_value",
    )


@pytest.mark.parametrize(
    "counts_raw, expected_out, include_zero_values",
    [
        ({" 1": 0, "0": 550}, {"0": 550}, False),
        ({"10": 479, "1 1": 13, "0 0 ": 496}, {"00": 496, "10": 479, "11": 13}, False),
        ({" 1": 474, "0": 550}, {"0": 550, "1": 474}, True),
        ({"10": 479, "1 1": 13, "0 0 ": 496}, {"00": 496, "01": 0, "10": 479, "11": 13}, True),
        (
            {"10 1": 586, "11  0  ": 139, "0 01": 496, "  010": 543, "11 1": 594},
            {
                "000": 0,
                "001": 496,
                "010": 543,
                "011": 0,
                "100": 0,
                "101": 586,
                "110": 139,
                "111": 594,
            },
            True,
        ),
    ],
)
def test_format_counts(counts_raw, expected_out, include_zero_values):
    """Test formatting of raw measurement counts."""
    counts_out = format_data(counts_raw, include_zero_values=include_zero_values)
    assert counts_out == expected_out  # check equivalance
    assert list(counts_out.items()) == list(expected_out.items())  # check ordering of keys


def test_format_counts_empty_input():
    """Test formatting of empty input."""
    counts = {}
    expected = {}
    assert format_data(counts) == expected


def test_normalize_different_key_lengths():
    """Test normalization of measurement counts with different key lengths."""
    measurements = [{"0": 10, "1": 15}, {"00": 5, "01": 8, "10": 12}]
    expected = [{"00": 10, "01": 15}, {"00": 5, "01": 8, "10": 12}]
    assert normalize_batch_bit_lengths(measurements) == expected


def test_normalize_same_key_lengths():
    """Test normalization of measurement counts with the same key lengths."""
    measurements = [{"00": 7, "01": 9}, {"10": 4, "11": 3}]
    expected = [{"00": 7, "01": 9}, {"10": 4, "11": 3}]
    assert normalize_batch_bit_lengths(measurements) == expected


def test_normalize_empty_input():
    """Test normalization of empty input."""
    measurements = []
    expected = []
    assert normalize_batch_bit_lengths(measurements) == expected


def test_normalize_empty_dicts():
    """Test normalization of empty dicts."""
    measurements = [{}, {}, {"00": 1, "11": 2}]
    expected = [{}, {}, {"00": 1, "11": 2}]
    assert normalize_batch_bit_lengths(measurements) == expected


def test_normalize_single_dict():
    """Test normalization of a single dict."""
    measurements = {"0": 2, "1": 3, "10": 4, "11": 1}
    expected = {"00": 2, "01": 3, "10": 4, "11": 1}
    assert normalize_bit_lengths(measurements) == expected


def test_get_counts_raises_for_no_measurements():
    """Test that an error is raised when no measurements are available."""
    data = GateModelResultData()
    with pytest.raises(ValueError) as excinfo:
        data.get_counts()
    assert str(excinfo.value) == "Counts data is not available."


def test_get_counts_from_cache_key():
    """Test that counts are retrieved from the cache key."""
    data = GateModelResultData(measurement_counts={"10": 2})
    assert all(value is None for value in data._cache.values())
    counts = data.get_counts()
    assert data._cache["bin_nz"] == counts
    data._cache["bin_nz"] = 42
    assert data.get_counts() == 42


class MockBatchResult:
    """Mock batch result for testing."""

    def get_counts(self):
        """Returns raw histogram data of the run."""
        return [{" 1": 0, "0": 550}, {" 1": 474, "0": 550}]


def test_batch_normalized_counts():
    """Test batch measurement counts."""
    result = MockBatchResult()
    raw_counts = result.get_counts()
    counts = normalize_data(raw_counts, include_zero_values=False)
    expected = [{"0": 550}, {"0": 550, "1": 474}]
    assert counts == expected


def test_decimal_get_counts():
    """Test decimal raw counts."""
    experiment = GateModelResultData(measurement_counts={"10": 2})
    result = Result("device_id", "job_id", True, experiment)
    counts = result.data.get_counts(decimal=True)
    expected = {2: 2}
    assert counts == expected


def test_result_data_experiment_type():
    """Test that result data classes return the correct experiment type."""
    gm_result_data = GateModelResultData()
    ahs_result_data = AhsResultData()
    assert gm_result_data.experiment_type == ExperimentType.GATE_MODEL
    assert ahs_result_data.experiment_type == ExperimentType.AHS


def test_to_dict_basic(gate_model_result_data):
    """Test the basic functionality of the to_dict method."""
    result_dict = gate_model_result_data.to_dict()

    assert isinstance(result_dict, dict)
    assert result_dict["shots"] == 175
    assert result_dict["num_measured_qubits"] == 2
    assert result_dict["measurement_counts"] == {"00": 100, "01": 50, "10": 25}
    assert "measurement_probabilities" in result_dict
    assert np.array_equal(result_dict["measurements"], np.array([[0, 0], [0, 1], [1, 0]]))


def test_to_dict_cache(gate_model_result_data):
    """Test that the to_dict method uses caching correctly."""
    assert gate_model_result_data._cache["to_dict"] is None
    gate_model_result_data.to_dict()

    assert gate_model_result_data._cache["to_dict"] is not None

    cached_result = gate_model_result_data._cache["to_dict"]
    result_dict = gate_model_result_data.to_dict()

    assert result_dict == cached_result


def test_to_dict_with_empty_measurements():
    """Test the to_dict method when there are no measurements provided."""
    measurement_counts = {"00": 100, "01": 50}
    result_data = GateModelResultData(measurement_counts=measurement_counts, measurements=None)

    result_dict = result_data.to_dict()

    assert result_dict["measurements"] is None
    assert result_dict["measurement_counts"] == {"00": 100, "01": 50}
    assert result_dict["shots"] == 150
    assert result_dict["num_measured_qubits"] == 2


def test_to_dict_probabilities(gate_model_result_data):
    """Test that the probabilities are correctly included in the dictionary."""
    result_dict = gate_model_result_data.to_dict()
    probabilities = result_dict["measurement_probabilities"]
    assert abs(sum(probabilities.values()) - 1) < 1e-6


def test_get_probabilities_from_cache(gate_model_result_data):
    """Test that probabilities are retrieved from cache if present."""
    assert gate_model_result_data._cache["prob_dec_nz"] is None
    calculated_probs = gate_model_result_data.get_probabilities(decimal=True)
    assert gate_model_result_data._cache["prob_dec_nz"] == calculated_probs
    mock_cached_probs = {}
    assert mock_cached_probs != calculated_probs
    gate_model_result_data._cache["prob_dec_nz"] = mock_cached_probs
    retrieved_probs = gate_model_result_data.get_probabilities(decimal=True)
    assert retrieved_probs == mock_cached_probs


@pytest.mark.parametrize(
    "precomputed_probs", ({"00": 0.4, "01": 0.6}, {"00": 0.25, "01": 0.25, "10": 0.25, "11": 0.25})
)
def test_precomputed_measurement_probs(precomputed_probs):
    """Tests that precomputed probabilities are returned from GateModelResultData"""
    measurement_probs = GateModelResultData(
        measurement_probabilities=precomputed_probs
    ).get_probabilities()
    assert precomputed_probs == measurement_probs


@pytest.mark.parametrize("precomputed_probs", ([0.1, 0.7, 0.2], True, False, 0.2, "probabilities"))
def test_precomputed_measurement_probs_not_dict(precomputed_probs):
    """Tests that ValueError is raised if 'measurement_probabilities' is not a dictionary."""
    with pytest.raises(ValueError, match="'measurement_probabilities' must be a dictionary."):
        GateModelResultData(measurement_probabilities=precomputed_probs).get_probabilities()


def test_to_dict_no_counts():
    """Test the to_dict method when measurement counts are not available."""
    result_data = GateModelResultData(measurement_counts=None, measurements=None)

    with pytest.raises(ValueError, match="Counts data is not available"):
        result_data.to_dict()


def test_ahs_shot_result_equality(shot_result):
    """Test equality of two AhsShotResult objects."""
    shot_result_2 = AhsShotResult(
        success=True,
        pre_sequence=np.array([1, 1, 1, 1, 0, 1, 1, 1, 1, 1, 1]),
        post_sequence=np.array([0, 0, 0, 1, 0, 1, 1, 0, 0, 1, 0]),
    )

    assert shot_result == shot_result_2

    shot_result_3 = AhsShotResult(
        success=True,
        pre_sequence=np.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1]),
        post_sequence=np.array([0, 0, 0, 1, 0, 1, 1, 0, 0, 1, 0]),
    )

    assert shot_result != shot_result_3
    assert shot_result != "not a shot result"


def test_ahs_shot_result_sequences_equal():
    """Test the _sequences_equal method."""
    seq1 = np.array([1, 1, 0])
    seq2 = np.array([1, 1, 0])

    assert AhsShotResult._sequences_equal(seq1, seq2)

    seq3 = np.array([0, 1, 0])
    assert not AhsShotResult._sequences_equal(seq1, seq3)
    assert AhsShotResult._sequences_equal(None, None)
    assert not AhsShotResult._sequences_equal(seq1, None)


def test_ahs_result_data_experiment_type(ahs_result_data):
    """Test that the experiment type is AHS."""
    assert ahs_result_data.experiment_type == ExperimentType.AHS


def test_ahs_result_data_measurements(ahs_result_data, shot_result):
    """Test the measurements property."""
    assert ahs_result_data.measurements == [shot_result]


def test_ahs_result_data_get_counts(ahs_result_data):
    """Test the get_counts method."""
    expected_counts = {"rrrgeggrrgr": 1}
    assert ahs_result_data.get_counts() == expected_counts


def test_ahs_result_data_to_dict(ahs_result_data):
    """Test the to_dict method of AhsResultData."""
    expected_dict = {
        "measurement_counts": {"rrrgeggrrgr": 1},
        "measurements": ahs_result_data.measurements,
    }

    assert ahs_result_data.to_dict() == expected_dict


def test_ahs_result_data_repr():
    """Test __repr__ method of AhsResultData."""
    ahs_result_data = AhsResultData(
        measurement_counts=[{"001010": 50, "001011": 50}], measurements=None
    )
    assert (
        repr(ahs_result_data)
        == "AhsResultData(measurement_counts=[{'001010': 50, '001011': 50}], measurements=None)"
    )


def test_ahs_result_data_no_measurement_counts():
    """Test to_dict and get_counts when there are no measurement counts."""
    result_data = AhsResultData(measurement_counts=None, measurements=None)

    assert result_data.get_counts() is None

    expected_dict = {
        "measurement_counts": None,
        "measurements": None,
    }
    assert result_data.to_dict() == expected_dict


def test_qir_sim_data_backend_version_property(qir_sim_data):
    """Test the backend_version property."""
    assert qir_sim_data.backend_version == "0.7.4"


def test_qir_sim_data_seed_property_with_value(qir_sim_data):
    """Test the seed property when a seed is provided."""
    assert qir_sim_data.seed == 42


def test_qir_sim_data_seed_property_with_none():
    """Test the seed property when no seed is provided."""
    result_data = QbraidQirSimulatorResultData(
        backend_version="v1.2.3",
        seed=None,
        measurement_counts={"00": 10, "01": 15},
    )

    assert result_data.seed is None


def test_qir_sim_data_inherited_measurement_counts_property(qir_sim_data):
    """Test that the inherited measurement_counts property is working correctly."""
    assert qir_sim_data.get_counts() == {"00": 10, "01": 15}


def test_nec_vector_annealer_result_data(nec_vector_annealer_result_data):
    """Test NECVectorAnnealerResultData."""
    assert nec_vector_annealer_result_data._solutions == [
        {
            "spin": {" x1": 0, " x2": 0, "x1": 0},
            "energy": 0,
            "time": 0.006517000030726194,
            "constraint": True,
            "memory_usage": 1.189453125,
        }
    ]


def test_nec_vector_annealer_result_data_to_dict(nec_vector_annealer_result_data):
    """Test NECVectorAnnealerResultData to_dict."""
    result_dict = nec_vector_annealer_result_data.to_dict()
    assert result_dict == {
        "solutions": [
            {
                "spin": {" x1": 0, " x2": 0, "x1": 0},
                "energy": 0,
                "time": 0.006517000030726194,
                "constraint": True,
                "memory_usage": 1.189453125,
            }
        ],
        "num_solutions": 1,
    }


def test_nec_vector_annealer_result_data_eq(nec_vector_annealer_result_data):
    """Test NECVectorAnnealerResultData equality."""
    assert nec_vector_annealer_result_data._solutions[0]["spin"] == {" x1": 0, " x2": 0, "x1": 0}
    assert nec_vector_annealer_result_data._solutions[0]["energy"] == 0
    assert nec_vector_annealer_result_data._solutions[0]["time"] == 0.006517000030726194
    assert nec_vector_annealer_result_data._solutions[0]["constraint"] is True
    assert nec_vector_annealer_result_data._solutions[0]["memory_usage"] == 1.189453125


def test_annealing_result_data(annealing_result_data):
    """Test AnnealingResultData."""
    assert annealing_result_data._solutions == [
        {
            "spin": {" x1": 0, " x2": 0, "x1": 0},
            "energy": 0,
            "time": 0.006517000030726194,
            "constraint": True,
            "memory_usage": 1.189453125,
        }
    ]


def test_annealing_result_from_dict():
    """Test AnnealingResultData from_dict."""
    solutions = [
        {
            "spin": {" x1": 0, " x2": 0, "x1": 0},
            "energy": 0,
            "time": 0.006517000030726194,
            "constraint": True,
            "memory_usage": 1.189453125,
        }
    ]
    result_data = AnnealingResultData.from_dict({"solutions": solutions})
    assert result_data._solutions == solutions


def test_annealing_result_data_get_result(annealing_result_data):
    """Test AnnealingResultData get_result."""
    result = annealing_result_data.solutions
    assert result == [
        {
            "spin": {" x1": 0, " x2": 0, "x1": 0},
            "energy": 0,
            "time": 0.006517000030726194,
            "constraint": True,
            "memory_usage": 1.189453125,
        }
    ]


def test_annealing_result_data_get_result_with_empty_solutions():
    """Test AnnealingResultData get_result with empty solutions."""
    result_data = AnnealingResultData(solutions=None)
    assert result_data.solutions is None


def test_annealing_result_data_get_result_experiment_type(annealing_result_data):
    """Test AnnealingResultData get_result experiment type."""
    assert annealing_result_data.experiment_type == ExperimentType.ANNEALING


def test_annealing_result_data_to_dict(annealing_result_data):
    """Test AnnealingResultData to_dict."""
    result_dict = annealing_result_data.to_dict()
    assert result_dict == {
        "solutions": [
            {
                "spin": {" x1": 0, " x2": 0, "x1": 0},
                "energy": 0,
                "time": 0.006517000030726194,
                "constraint": True,
                "memory_usage": 1.189453125,
            }
        ],
        "num_solutions": 1,
    }


def test_normalize_batch_decimal_counts():
    """Test normalization of batch measurement counts with decimal=True."""
    counts = [{"00": 10, "01": 15}, {"10": 20, "11": 25}]
    expected = [{0: 10, 1: 15}, {2: 20, 3: 25}]
    result = normalize_data(counts, decimal=True)
    assert result == expected


@pytest.mark.parametrize(
    "counts", [{0: 10, 1: 15, 3: 25}, {"0": 10, "1": 15, "3": 25}, {"00": 10, "01": 15, "11": 25}]
)
def test_format_counts_include_zero_values_decimal(counts):
    """Test format counts include_zero_values option in decimal form."""
    expected = {0: 10, 1: 15, 2: 0, 3: 25}
    result = format_data(counts, include_zero_values=True, decimal=True)
    assert result == expected


def test_format_value_string(result_instance):
    """Test _format_value with a string input."""
    assert result_instance._format_value("hello") == "'hello'"


def test_format_value_enum(result_instance):
    """Test _format_value with an Enum input."""
    assert result_instance._format_value(ExperimentType.GATE_MODEL) == "GATE_MODEL"


def test_format_value_datetime(result_instance):
    """Test _format_value with a datetime input."""
    test_datetime = datetime.datetime(2023, 1, 1, 12, 0)
    assert result_instance._format_value(test_datetime) == "2023-01-01T12:00:00Z"


def test_format_value_dict(result_instance):
    """Test _format_value with a dict input."""
    test_dict = {"key1": "value1", "key2": "value2"}
    assert result_instance._format_value(test_dict) == "{key1: 'value1', key2: 'value2'}"


def test_format_value_dict_with_depth(result_instance):
    """Test _format_value with a nested dict input and depth limit."""
    test_dict = {"key1": {"nested_key": "nested_value"}}
    assert result_instance._format_value(test_dict, depth=2) == "{...}"


def test_format_value_list(result_instance):
    """Test _format_value with a list input."""
    test_list = ["item1", "item2", "item3"]
    assert result_instance._format_value(test_list) == "['item1', 'item2', 'item3']"


def test_format_value_list_with_depth(result_instance):
    """Test _format_value with a nested list and depth limit."""
    test_list = [["item1", "item2"], ["item3", "item4"]]
    assert result_instance._format_value(test_list, depth=2) == "[...]"


def test_format_value_dict_with_openqasm(result_instance):
    """Test _format_value with a dict with openQasm key."""
    test_dict = {"openQasm": "OPENQASM 2.0;"}
    assert result_instance._format_value(test_dict) == "{openQasm: '...'}"


def test_format_value_catch_all(result_instance):
    """Test _format_value with for catch-all repr."""
    assert result_instance._format_value(42) == "42"


def test_repr(result_instance):
    """Test __repr__ method of Result class."""
    expected_repr = (
        "Result(\n"
        "  device_id=test_device,\n"
        "  job_id=test_job,\n"
        "  success=True,\n"
        "  data=GateModelResultData("
        "measurement_counts=None, "
        "measurements=None, "
        "measurement_probabilities=None"
        ")"
    )
    assert repr(result_instance).startswith(expected_repr)


@pytest.mark.parametrize(
    "counts1, counts2, expected",
    [
        ({"00": 5, "01": 3}, {"00": 5, "01": 3}, True),
        ({"00": 5, "01": 3}, {"00": 4, "01": 3}, False),
        (None, None, True),
        ({"00": 5}, None, False),
    ],
)
def test_ahs_result_data_measurement_counts_equality(counts1, counts2, expected):
    """Test equality of AhsResultData objects with different measurement counts."""
    shot = AhsShotResult(success=True, pre_sequence=np.array([1, 2]), post_sequence=None)
    result1 = AhsResultData(measurement_counts=counts1, measurements=[shot])
    result2 = AhsResultData(measurement_counts=counts2, measurements=[shot])
    assert (result1 == result2) == expected


@pytest.mark.parametrize(
    "measurements1, measurements2, expected",
    [
        (None, None, True),
        ([None], None, False),
        ([None], [None], True),
        ([], [], True),
        ([None], [AhsShotResult(success=True)], False),
    ],
)
def test_ahs_result_data_measurements_none_equality(measurements1, measurements2, expected):
    """Test equality of AhsResultData objects with different measurements."""
    result1 = AhsResultData(measurement_counts=None, measurements=measurements1)
    result2 = AhsResultData(measurement_counts=None, measurements=measurements2)
    assert (result1 == result2) == expected


@pytest.mark.parametrize(
    "measurements1, measurements2, expected",
    [
        ([AhsShotResult(success=True)], [AhsShotResult(success=True)], True),
        ([AhsShotResult(success=True)], [AhsShotResult(success=False)], False),
        ([AhsShotResult(success=True)], [], False),
        (
            [AhsShotResult(success=True, pre_sequence=np.array([1, 2]))],
            [AhsShotResult(success=True, pre_sequence=np.array([1, 3]))],
            False,
        ),
    ],
)
def test_ahs_result_data_measurements_list_equality(measurements1, measurements2, expected):
    """Test equality of AhsResultData objects with different measurements."""
    result1 = AhsResultData(measurement_counts=None, measurements=measurements1)
    result2 = AhsResultData(measurement_counts=None, measurements=measurements2)
    assert (result1 == result2) == expected


def test_ahs_result_data_different_class_equality():
    """Test equality of AhsResultData objects with different classes."""
    result1 = AhsResultData(measurement_counts=None, measurements=None)
    assert result1 != "not an AhsResultData"


def test_ahs_result_data_from_dict_measurements_not_list():
    """Test that ValueError is raised if 'measurements' is not a list or None."""
    invalid_data = {
        "measurements": "not_a_list",
        "measurement_counts": {"00": 5, "01": 3},
    }
    with pytest.raises(ValueError, match="'measurements' must be a list or None."):
        AhsResultData.from_dict(invalid_data)


def test_ahs_result_data_from_dict_measurements_items_not_dict():
    """Test that ValueError is raised if any item in 'measurements' is not a dictionary."""
    invalid_data = {
        "measurements": ["not_a_dict"],
        "measurement_counts": {"00": 5, "01": 3},
    }
    with pytest.raises(ValueError, match="Each item in 'measurements' must be a dictionary."):
        AhsResultData.from_dict(invalid_data)


def test_ahs_result_data_from_dict_valid_data():
    """Test that valid data does not raise an exception."""
    valid_data = {
        "measurements": [{"success": True, "pre_sequence": [1, 2], "post_sequence": [3, 4]}],
        "measurement_counts": {"00": 5, "01": 3},
    }
    result = AhsResultData.from_dict(valid_data)
    assert result.measurements is not None
    assert len(result.measurements) == 1
    assert result.measurements[0].success is True
    assert result.get_counts() == {"00": 5, "01": 3}


def test_distribute_counts_valid_input():
    """Test distribute_counts with valid input."""
    probs = {0: 0.86, 1: 0.14}
    shots = 10
    result = distribute_counts(probs, shots)
    assert result == {0: 9, 1: 1}, "Counts do not match expected values."


def test_distribute_counts_probabilities_close_to_1():
    """Test distribute_counts with valid input (sum of probabilities close to 1)."""
    probs = {0: 0.888888888, 1: 0.111111111}
    shots = 10
    result = distribute_counts(probs, shots)
    assert result == {0: 9, 1: 1}, "Counts do not match expected values."


def test_distribute_counts_probabilities_close_to_1_rel_tol():
    """Test distribute_counts with valid input (sum of probabilities close to 1)
    given rel_tol=1e-7."""
    probs = {
        "0": 0.9869999885559082,
        "1": 0.004000000189989805,
        "2": 0.004999999888241291,
        "3": 0.004000000189989805,
    }
    shots = 100000
    result = distribute_counts(probs, shots)
    assert result == {
        "0": 98700,
        "1": 400,
        "2": 500,
        "3": 400,
    }, "Counts do not match expected values."


def test_distribute_counts_probabilities_not_sum_to_1():
    """Test distribute_counts with probabilities that do not sum to 1."""
    probs = {0: 0.9, 1: 0.2}
    shots = 10
    with pytest.warns(UserWarning, match="Probabilities do not sum to 1."):
        distribute_counts(probs, shots)


def test_distribute_counts_negative_probability():
    """Test distribute_counts with negative probability."""
    probs = {0: -0.1, 1: 1.1}
    shots = 10
    with pytest.raises(ValueError, match="Probabilities must be between 0 and 1."):
        distribute_counts(probs, shots)


def test_distribute_counts_negative_shots():
    """Test distribute_counts with negative number of shots."""
    probs = {0: 0.86, 1: 0.14}
    shots = -10
    with pytest.raises(ValueError, match="Number of shots must be non-negative."):
        distribute_counts(probs, shots)


def test_distribute_counts_zero_shots():
    """Test distribute_counts with zero shots."""
    probs = {0: 0.86, 1: 0.14}
    shots = 0
    result = distribute_counts(probs, shots)
    assert result == {0: 0, 1: 0}, "Counts do not match expected values for zero shots."


def test_distribute_counts_large_input():
    """Test distribute_counts with large input."""
    probs = {0: 0.5, 1: 0.5}
    shots = 1000000
    result = distribute_counts(probs, shots)
    assert sum(result.values()) == shots, "Counts do not sum to the number of shots."
    assert (
        result[0] == 500000 and result[1] == 500000
    ), "Counts do not distribute evenly for equal probabilities."


def test_distribute_counts_edge_case_probabilities():
    """Test distribute_counts with edge case probabilities."""
    probs = {0: 0.999999, 1: 0.000001}
    shots = 10
    result = distribute_counts(probs, shots)
    assert sum(result.values()) == shots, "Counts do not sum to the number of shots."
    assert result[0] == 10 and result[1] == 0, "Counts do not match expected edge case behavior."


def test_distribute_counts_diff_non_zero():
    """Test distribute_counts with probabilities that are very close but not equal."""
    probs = {0: 0.3333, 1: 0.3333, 2: 0.3334}
    shots = 10
    result = distribute_counts(probs, shots)
    assert sum(result.values()) == shots, "Counts do not sum to the number of shots."
    assert result[0] + result[1] + result[2] == shots, "Counts adjustment did not work as expected."



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unit test for quantum program registry for QIR

"""
from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from qbraid.programs import ExperimentType, ProgramSpec, unregister_program_type

if TYPE_CHECKING:
    import pyqir
    from pyqir import Module

pyqir = pytest.importorskip("pyqir")


@pytest.fixture
def pyqir_bell() -> Module:
    """Returns a QIR bell circuit with measurement over two qubits."""
    bell = pyqir.SimpleModule("test_qir_bell", num_qubits=2, num_results=2)
    qis = pyqir.BasicQisBuilder(bell.builder)

    qis.h(bell.qubits[0])
    qis.cx(bell.qubits[0], bell.qubits[1])
    qis.mz(bell.qubits[0], bell.results[0])
    qis.mz(bell.qubits[1], bell.results[1])

    return bell._module


@pytest.fixture
def pyqir_spec() -> ProgramSpec:
    """Returns a ProgramSpec for the QIR bell circuit."""
    return ProgramSpec(pyqir.Module, alias="pyqir", serialize=lambda module: module.bitcode)


def test_load_program_pyqir(pyqir_bell: Module, pyqir_spec: ProgramSpec):
    """Test program spec to IR conversion for a QIR program."""
    program_ir = pyqir_spec.serialize(pyqir_bell)
    assert isinstance(program_ir, bytes)
    assert program_ir == pyqir_bell.bitcode


def test_program_spec_specify_experiment_type():
    """Test specifying the experiment type for a program spec."""
    try:
        spec = ProgramSpec(bytes, alias="llvm", experiment_type=ExperimentType.OTHER)
        assert spec.experiment_type == ExperimentType.OTHER
    finally:
        unregister_program_type("llvm", raise_error=False)



# Copyright (C) 2025 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unit tests for loading jobs using entrypoints

"""

import pytest

from qbraid.runtime import (
    PROVIDERS,
    JobLoaderError,
    ProviderLoaderError,
    QbraidProvider,
    get_providers,
    load_job,
    load_provider,
)

from ._resources import JOB_DATA_QIR


@pytest.fixture
def job_id():
    """Mock job data for testing"""
    return JOB_DATA_QIR["qbraidJobId"]


def test_load_job(mock_client, job_id):
    """Test loading a job using entrypoints"""
    job = load_job(job_id, "qbraid", client=mock_client)
    assert job.id == job_id


def test_load_job_error(job_id):
    """Test that JobLoaderError is raised when loading a job fails."""
    provider = "fake_provider"

    with pytest.raises(
        JobLoaderError,
        match=f"Error loading QuantumJob sub-class for provider '{provider}'.",
    ):
        load_job(job_id, provider)


def test_get_providers():
    """Test getting all available providers."""
    providers = get_providers()
    assert providers == PROVIDERS == ["aws", "azure", "ibm", "ionq", "oqc", "qbraid"]


def test_load_provider(mock_client):
    """Test loading a provider using entrypoints"""
    provider = load_provider("qbraid", client=mock_client)
    assert isinstance(provider, QbraidProvider)


def test_load_provider_error():
    """Test that ProviderLoaderError is raised when loading a provider fails."""
    provider_name = "fake_provider"

    with pytest.raises(
        ProviderLoaderError,
        match=f"Error loading QuantumProvider sub-class for provider '{provider_name}'.",
    ):
        load_provider(provider_name)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining Qiskit OpenQASM conversions

"""
from __future__ import annotations

from typing import TYPE_CHECKING

from qbraid_core._import import LazyLoader

from qbraid.passes.qasm.compat import add_stdgates_include, insert_gate_def, replace_gate_names
from qbraid.transpiler.annotations import weight

qiskit_qasm3 = LazyLoader("qiskit_qasm3", globals(), "qiskit.qasm3")


if TYPE_CHECKING:
    import qiskit as qiskit_

    from qbraid.programs.typer import Qasm3StringType


def transform_notation(qasm3: str) -> str:
    """
    Process an OpenQASM 3 program that was generated by
    an external tool to make it compatible with Qiskit.

    """
    replacements = {
        "cnot": "cx",
        "si": "sdg",
        "ti": "tdg",
        "v": "sx",
        "vi": "sxdg",
        "phaseshift": "p",
        "cphaseshift": "cp",
    }

    qasm3 = replace_gate_names(qasm3, replacements)
    qasm3 = add_stdgates_include(qasm3)
    qasm3 = insert_gate_def(qasm3, "iswap")
    qasm3 = insert_gate_def(qasm3, "sxdg")

    return qasm3


@weight(1)
def qasm3_to_qiskit(qasm: Qasm3StringType) -> qiskit_.QuantumCircuit:
    """Convert QASM 3.0 string to a Qiskit QuantumCircuit representation.

    Args:
        qasm (str): A string in QASM 3.0 format.

    Returns:
        qiskit.QuantumCircuit: A QuantumCircuit object representing the input QASM 3.0 string.
    """
    try:
        return qiskit_qasm3.loads(qasm)
    except qiskit_qasm3.QASM3ImporterError:
        pass

    qasm = transform_notation(qasm)

    return qiskit_qasm3.loads(qasm)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module for generaing random quantum circuits and unitary matrices
for testing.

.. currentmodule:: qbraid.interface.random

Functions
----------

.. autosummary::
   :toctree: ../stubs/

   random_circuit
   random_unitary_matrix

"""
from .random import random_circuit, random_unitary_matrix

__all__ = ["random_circuit", "random_unitary_matrix"]



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

# pylint: disable=redefined-outer-name

"""
Unit tests for quantum jobs functions and data types

"""
from unittest.mock import patch

import pytest

from qbraid.programs import ExperimentType
from qbraid.runtime import QuantumJob
from qbraid.runtime.enums import JobStatus
from qbraid.runtime.exceptions import (
    DeviceProgramTypeMismatchError,
    JobStateError,
    QbraidRuntimeError,
    ResourceNotFoundError,
)
from qbraid.runtime.native.job import QbraidJob

from ._resources import JOB_DATA_QIR


def get_expected_message(program, expected_type, experiment_type):
    """Generate the expected error message dynamically."""
    if program is None:
        program_type = "NoneType"
    else:
        program_type = type(program).__name__
    return (
        f"Incompatible program type: '{program_type}'. "
        f"Experiment type '{experiment_type}' requires a program of type '{expected_type}'."
    )


class MockProgram:
    """Mock class for testing DeviceProgramTypeMismatchError."""


@pytest.mark.parametrize(
    "program, expected_type, experiment_type",
    [
        (MockProgram(), "QuantumProgram", "Compute"),
        (object(), "QuantumProgram", "Compute"),
        (MockProgram(), "MockProgram", "Compute"),
        (None, "QuantumProgram", "Compute"),
    ],
)
def test_device_program_type_mismatch_error(program, expected_type, experiment_type):
    """Test DeviceProgramTypeMismatchError with various scenarios."""
    expected_message = get_expected_message(program, expected_type, experiment_type)
    with pytest.raises(DeviceProgramTypeMismatchError) as excinfo:
        raise DeviceProgramTypeMismatchError(program, expected_type, experiment_type)

    assert str(excinfo.value) == expected_message


class FakeDevice:
    """Fake device class for testing."""

    def __init__(self, name, client):
        self.name = name
        self.client = client


def test_qbraid_job_device():
    """Test setting and getting device."""
    job = QbraidJob("test_job")
    with pytest.raises(ResourceNotFoundError):
        job.device  # pylint: disable=pointless-statement

    fake_device = FakeDevice("test_device", "test_client")

    job2 = QbraidJob("test_job", device=fake_device)
    assert job2.device == fake_device
    assert job2.client == "test_client"


class FakeQbraidJob(QbraidJob):
    """Fake QbraidJob class for testing."""

    def __init__(self, job_id, terminal, device=None, client=None):
        super().__init__(job_id, device)
        self._terminal = terminal
        self._client = client

    def is_terminal_state(self):
        """Check if the job is in a terminal state."""
        return self._terminal


class FakeQbraidClient:
    """Fake QbraidClient class for testing"""

    def __init__(
        self,
        *args,
        init_status="FAILED",
        init_status_text="Some error message",
        fail_cancel=False,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self._job_data = JOB_DATA_QIR.copy()
        self._job_data["status"] = init_status
        self._job_data["statusText"] = init_status_text
        self._fail_cancel = fail_cancel

    def get_job(self, job_id):  # pylint: disable=unused-argument
        """Get the job data."""
        status_data_copy = self._job_data.copy()
        if self._job_data["status"] == "CANCELLING":
            self._job_data["statusText"] = ""
            self._job_data["status"] = "CANCELLED"
        return status_data_copy

    def cancel_job(self, qbraid_id):  # pylint: disable=unused-argument
        """Cancel a job."""
        if self._fail_cancel:
            self._job_data["status"] = "COMPLETED"
            self._job_data["statusText"] = ""
        else:
            self._job_data["status"] = "CANCELLING"
            self._job_data["statusText"] = ""


@pytest.fixture
def mock_client():
    """Return a mock QbraidClient object."""
    return FakeQbraidClient()


@pytest.fixture
def mock_device(mock_client):
    """Return a mock QbraidDevice object."""
    return FakeDevice("test_device", client=mock_client)


@pytest.fixture
def mock_job(mock_device, mock_client):
    """Return a mock QbraidJob object."""
    return QbraidJob("test_job", device=mock_device, client=mock_client)


def test_cancel_job():
    """Test cancelling a job."""
    terminal_job = FakeQbraidJob("test_job", True)
    with pytest.raises(JobStateError):
        terminal_job.cancel()

    nonterminal_job = FakeQbraidJob(
        "test_job", False, client=FakeQbraidClient(init_status="RUNNING", init_status_text="")
    )
    assert nonterminal_job.cancel() is None

    nonterminal_job = FakeQbraidJob(
        "test_job",
        False,
        client=FakeQbraidClient(init_status="RUNNING", init_status_text="", fail_cancel=True),
    )
    with pytest.raises(QbraidRuntimeError):
        nonterminal_job.cancel()


class MockQuantumJob(QuantumJob):
    """Mock class for testing QuantumJob functions."""

    def __init__(self, job_id, device=None, **kwargs):
        super().__init__(job_id, device, **kwargs)
        self._status = "PENDING"

    def status(self):
        return self._status

    def cancel(self):
        pass

    def result(self):
        pass


@pytest.fixture
def quantum_job():
    """Return a mock QuantumJob object"""
    job = MockQuantumJob(job_id="test_job_id")
    return job


@pytest.mark.asyncio
async def test_async_wait_for_final_state_success(quantum_job):
    """Ensures the async method completes when the job reaches a terminal state."""
    with patch.object(quantum_job, "is_terminal_state", side_effect=[False, False, True]):
        await quantum_job._wait_for_final_state(timeout=1, poll_interval=0.1)


@pytest.mark.asyncio
async def test_async_wait_for_final_state_timeout(quantum_job):
    """Ensures the async method raises TimeoutError if the job never completes."""
    with patch.object(quantum_job, "is_terminal_state", return_value=False):
        with pytest.raises(TimeoutError):
            await quantum_job._wait_for_final_state(timeout=0.2, poll_interval=0.1)


@pytest.mark.asyncio
async def test_async_result_success(quantum_job):
    """Test that async_result returns the job result after reaching terminal state."""
    mock_result = "expected_result"

    with patch.object(quantum_job, "is_terminal_state", side_effect=[False, True]):
        with patch.object(quantum_job, "result", return_value=mock_result):
            result = await quantum_job.async_result(timeout=1, poll_interval=0.1)
            assert result == mock_result


@pytest.mark.asyncio
async def test_async_result_timeout(quantum_job):
    """Test that async_result raises TimeoutError if job never reaches terminal state."""
    with patch.object(quantum_job, "is_terminal_state", return_value=False):
        with pytest.raises(TimeoutError):
            await quantum_job.async_result(timeout=0.2, poll_interval=0.1)


def test_wait_for_final_state_success(quantum_job):
    """Mocking the status to change to a final state after some time"""
    with patch.object(quantum_job, "is_terminal_state", side_effect=[False, False, True]):
        quantum_job.wait_for_final_state(timeout=1, poll_interval=0.1)


def test_wait_for_final_state_timeout(quantum_job):
    """Mocking the status to never change to a final state"""
    with patch.object(quantum_job, "is_terminal_state", return_value=False):
        with pytest.raises(TimeoutError):
            quantum_job.wait_for_final_state(timeout=0.2, poll_interval=0.1)


def test_invalid_job_status_value():
    """Test that an invalid status value raises a ValueError."""
    with pytest.raises(ValueError, match="Invalid status value: INVALID_STATUS"):
        JobStatus._get_default_message("INVALID_STATUS")


def test_set_job_status_message():
    """Test that a custom status message can be set."""
    status = JobStatus.QUEUED
    assert status.default_message == "job is queued"
    assert status.status_message is None

    custom_status_message = "Custom status message"
    status.set_status_message(custom_status_message)
    assert status.status_message == custom_status_message
    assert repr(status) == f"<QUEUED: '{custom_status_message}'>"


def test_set_job_status_from_status_text(mock_job):
    """Test that a status message is set correctly from statusText job data received from client."""
    status = mock_job.status()
    assert status == JobStatus.FAILED
    assert status.status_message == "Some error message"


def test_job_status_enum_call_method():
    """Test that the JobStatus enum can be called as a function."""
    status = JobStatus.QUEUED
    new_status = status()
    assert new_status == status


def test_job_get_result_cls_raises_for_mismatch_expt_type():
    """Test that get_result_data_cls raises a ValueError for unsupported experiment type."""
    with pytest.raises(
        ValueError, match="Unsupported device_id 'aws_sv1' or experiment_type 'PHOTONIC'"
    ):
        QbraidJob.get_result_data_cls("aws_sv1", ExperimentType.PHOTONIC)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module for runtime quantum program transformations.

.. currentmodule:: qbraid.passes

Exceptions
-----------

.. autosummary::
   :toctree: ../stubs/

   CompilationError
   QasmDecompositionError

"""
from .exceptions import CompilationError, QasmDecompositionError

__all__ = ["CompilationError", "QasmDecompositionError"]

_lazy_mods = ["qasm"]


def __getattr__(name):
    if name in _lazy_mods:
        import importlib  # pylint: disable=import-outside-toplevel

        module = importlib.import_module(f".{name}", __name__)
        globals()[name] = module
        return module
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


def __dir__():
    return sorted(__all__ + _lazy_mods)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module for transpiling quantum programs between different quantum programming languages

"""

from __future__ import annotations

import warnings
from copy import deepcopy
from typing import TYPE_CHECKING, Any, Callable, Optional

from qbraid_core._import import LazyLoader

from qbraid._logging import logger
from qbraid.programs import QPROGRAM_ALIASES
from qbraid.programs.alias_manager import (
    _get_program_type_alias,
    get_program_type_alias,
)

from .exceptions import (
    ConversionPathNotFoundError,
    NodeNotFoundError,
    ProgramConversionError,
)
from .graph import ConversionGraph, _get_path_from_bound_methods

if TYPE_CHECKING:
    import qbraid.programs


def _warn_if_unsupported(program_type, program_direction):
    if program_type not in QPROGRAM_ALIASES:
        warnings.warn(
            f"Converting {program_direction} unsupported program type '{program_type}'.",
            UserWarning,
        )


def _format_exception(err: Exception) -> str:
    return f"{type(err).__name__}: {str(err)}\n"


def transpile(
    program: qbraid.programs.QPROGRAM,
    target: str,
    conversion_graph: Optional[ConversionGraph] = None,
    max_path_attempts: int = 3,
    max_path_depth: Optional[int] = None,
    **kwargs,
) -> qbraid.programs.QPROGRAM:
    """
    Transpile a quantum program to a target language using a conversion graph.
    This function attempts to find a conversion path from the program's current
    format to the target format. It can limit the search to a certain number of
    attempts and path depths.

    Args:
        program (qbraid.programs.QPROGRAM): The quantum program to transpile.
        target (str): The target language to transpile to.
        conversion_graph (Optional[ConversionGraph]): The graph representing available conversions.
            If None, a default graph is used. Defaults to None.
        max_path_attempts (int): The maximum number of conversion paths to attempt before raising an
            exception. This is useful to avoid excessive computations when multiple paths are
            available. Defaults to 3.
        max_path_depth (Optional[int]): The maximum depth of conversions within a given path to
            allow. For example, a path with a depth of 2 would be ['cirq' -> 'qasm2' -> 'qiskit'],
            whereas a depth  of 1 would be a direct conversion ['cirq' -> 'braket']. Defaults
            to None, i.e. no limit set on the path depth.

    Returns:
        qbraid.programs.QPROGRAM: The transpiled quantum program.

    Raises:
        NodeNotFoundError: If the target or source package is not in the ConversionGraph.
        ConversionPathNotFoundError: If no path is available to conversion between the
            source and target packages.
        ProgramConversionError: If the conversion fails through all attempted paths.
    """
    graph = conversion_graph or ConversionGraph(**kwargs)
    graph_type = "Default" if conversion_graph is None else "Provided"

    if not graph.has_node(target):
        raise NodeNotFoundError(graph_type, target, graph.nodes())

    source = _get_program_type_alias(program)

    if not graph.has_node(source):
        raise NodeNotFoundError(graph_type, source, graph.nodes())

    if not graph.has_path(source, target):
        raise ConversionPathNotFoundError(source, target)

    if source == target:
        return program

    _warn_if_unsupported(source, "from")
    _warn_if_unsupported(target, "to")

    paths = graph.find_top_shortest_conversion_paths(source, target, top_n=max_path_attempts)

    if max_path_depth is not None:
        paths = [path for path in paths if len(path) <= max_path_depth]
        if len(paths) == 0:
            raise ConversionPathNotFoundError(source, target, max_path_depth)

    error_messages = []

    for path in paths:
        path_details = _get_path_from_bound_methods(path)
        try:
            temp_program = deepcopy(program)
        except (RecursionError, TypeError) as err:
            logger.info(
                "Deepcopy failed due to a %s, likely caused by the internal structure of "
                "the %s object. Continuing execution, but any subsequent errors during "
                "transpilation may be unclear or misleading due to potential side effects.",
                type(err).__name__,
                type(program),
            )
            temp_program = program
        try:
            for convert_func in path:
                try:
                    temp_program = convert_func(temp_program)
                except Exception as err:  # pylint: disable=broad-exception-caught
                    alias = get_program_type_alias(temp_program, safe=True)

                    if alias == "cirq":
                        cirq_qasm_import = LazyLoader(
                            "cirq_qasm_import", globals(), "cirq.contrib.qasm_import"
                        )
                        qasm: str = temp_program.to_qasm()  # type: ignore[attr-defined]
                        temp_program = cirq_qasm_import.circuit_from_qasm(qasm)
                        temp_program = convert_func(temp_program)  # Retry conversion
                    else:
                        error_detail = (
                            f"Conversion {path_details} failed due to "
                            f"exception raised while converting from '{alias}'."
                        )
                        error_messages.append(error_detail)
                        error_messages.append(_format_exception(err))
                        raise

            logger.info("Successfully transpiled using conversions: %s", path_details)
            return temp_program
        except Exception as err:  # pylint: disable=broad-exception-caught
            logger.info("Failed to transpile using conversions: %s", path_details)
            formatted_error = _format_exception(err)
            if len(error_messages) == 0 or error_messages[-1] != formatted_error:
                error_messages.append(formatted_error)
            continue

    raise ProgramConversionError(
        f"Failed to convert '{source}' to '{target}'"
        + (
            " due to the following error(s):\n\n" + "\n".join(error_messages)
            if error_messages
            else "."
        )
    )


def chain_calls(func: Callable[[Any, Any], Any], initial_value, *args, **kwargs) -> Any:
    """
    Apply a function iteratively over a sequence of arguments.

    Args:
        func (Callable[[Any, Any], Any]): The function to apply.
        initial_value (Any): The initial input to the first call.
        *args (Any): The sequence of arguments for subsequent calls.
        **kwargs (Any): Additional keyword arguments for the function.

    Returns:
        Any: The result of the final call.
    """
    result = initial_value
    for arg in args:
        result = func(result, arg, **kwargs)
    return result


def translate(program, *targets, **kwargs):
    """
    Chain multiple transpile calls on a program.

    Args:
        program: The quantum program to transpile.
        *targets: A sequence of target formats.

    Returns:
        The final transpiled program.
    """
    return chain_calls(transpile, program, *targets, **kwargs)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining :py:class:`ResultData` subclasses for runtime jobs that are
managed natively through qBraid.

"""
from __future__ import annotations

import base64
from typing import TYPE_CHECKING, Any, Optional

from qbraid.exceptions import QbraidError
from qbraid.programs.typer import Qasm2StringType
from qbraid.runtime.result_data import AnnealingResultData, GateModelResultData

if TYPE_CHECKING:
    import flair_visual.animation.runtime.qpustate
    import flair_visual.simulation_result
    import pandas as pd


class QuEraQasmSimulatorResultData(GateModelResultData):
    """Class for storing and accessing the results of a QuEra QASM simulator job."""

    def __init__(
        self, backend: str, quera_simulation_result: Optional[dict[str, Any]] = None, **kwargs
    ):
        """Create a new QuEraSimulatorResultData instance."""
        super().__init__(**kwargs)
        self._backend = backend
        self._quera_simulation_result: Optional[
            flair_visual.simulation_result.QuEraSimulationResult
        ] = (
            self._initialize_simulation_result(quera_simulation_result)
            if quera_simulation_result
            else None
        )

    @property
    def backend(self) -> str:
        """Returns the backend used."""
        return self._backend

    def _initialize_simulation_result(
        self, value: dict[str, Any]
    ) -> flair_visual.simulation_result.QuEraSimulationResult:
        """Initialize the QuEra simulation result during construction."""
        try:
            # pylint: disable-next=import-outside-toplevel
            from flair_visual.simulation_result import QuEraSimulationResult

            return QuEraSimulationResult.from_json(value)
        except ImportError as err:  # pragma: no cover
            raise QbraidError(
                "Missing required package 'flair-visual'. "
                "Install with 'pip install flair-visual' or 'pip install qbraid[quera]'."
            ) from err

    @property
    def quera_simulation_result(self) -> flair_visual.simulation_result.QuEraSimulationResult:
        """Returns the QuEra simulation result."""
        if self._quera_simulation_result is None:
            raise ValueError("The simulation result is not available.")
        return self._quera_simulation_result

    @property
    def flair_visual_version(self) -> str:
        """Returns the version of the flair visualizer used."""
        return self.quera_simulation_result.flair_visual_version

    def get_qpu_state(self) -> flair_visual.animation.runtime.qpustate.AnimateQPUState:
        """Returns the state of the QPU atoms used in the simulation."""
        return self.quera_simulation_result.atom_animation_state

    def get_logs(self) -> pd.DataFrame:
        """Returns the logs generated during the simulation as a pandas DataFrame."""
        return self.quera_simulation_result.logs


class QbraidQirSimulatorResultData(GateModelResultData):
    """Class for storing and accessing the results of a qBraid QIR simulator job."""

    def __init__(self, backend_version: str, seed: Optional[int] = None, **kwargs):
        """Create a new QbraidQirSimulatorResultData instance."""
        super().__init__(**kwargs)
        self._backend_version = backend_version
        self._seed = seed

    @property
    def backend_version(self) -> str:
        """Returns the version of the simulator backend used."""
        return self._backend_version

    @property
    def seed(self) -> Optional[int]:
        """Returns the seed used for the simulation."""
        return self._seed


class Equal1SimulatorResultData(GateModelResultData):
    """Class for storing and accessing the results of an Equal1 simulator job."""

    def __init__(self, compiled_output: str, **kwargs):
        """Create a new Equal1SimulatorResultData instance."""
        super().__init__(**kwargs)
        self._compiled_output = self._decode_base64(compiled_output)

    @staticmethod
    def _decode_base64(value: str) -> str:
        return base64.b64decode(value.encode("utf-8")).decode("utf-8")

    @property
    def compiled_output(self) -> Qasm2StringType:
        """Returns the compiled output of the simulation."""
        return self._compiled_output


class NECVectorAnnealerResultData(AnnealingResultData):
    """Class for storing and accessing the results of an NEC job."""



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module containing models for schema-conformant Results.

"""
from __future__ import annotations

import copy
import datetime
from enum import Enum
from typing import Any, Generic

from qbraid_core import deprecated
from qbraid_core.system.generic import _datetime_to_str

from .result_data import ResultDataType


class Result(Generic[ResultDataType]):
    """Represents the results of a quantum job. This class is intended
    to be initialized by a QuantumJob class.

    Args:
        device_id (str): The ID of the device that executed the job.
        job_id (str or int): The ID of the job.
        success (bool): Whether the job was successful.
        data (ResultData): The result of the job.
        **details: Additional metadata about the job results

    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        device_id: str,
        job_id: str | int,
        success: bool,
        data: ResultDataType,
        **kwargs,
    ):
        """Create a new Result object."""
        self.device_id = device_id
        self.job_id = job_id
        self.success = success
        self._data = data
        self._details = kwargs or {}

    @property
    def data(self) -> ResultDataType:
        """Returns the result of the job."""
        return self._data

    @property
    def details(self) -> dict[str, Any]:
        """Returns the result of the job."""
        return self._details

    @deprecated("Use 'Result.data.get_counts()' instead.")
    def measurement_counts(self, *args, **kwargs) -> Any:
        """Returns the measurement counts of the job."""
        return self.data.get_counts(*args, **kwargs)

    @deprecated("Use 'Result.data.measurements' instead.")
    def measurements(self) -> Any:
        """Returns the measurements of the job."""
        return self.data.measurements

    def _format_value(self, value: Any, depth: int = 0, max_depth: int = 2) -> str:
        """Helper function to format nested values with a depth limit."""
        if isinstance(value, str):
            return f"'{value}'"
        if isinstance(value, Enum):
            return f"{value.name}"
        if isinstance(value, datetime.datetime):
            return _datetime_to_str(value)
        if isinstance(value, dict):
            if depth >= max_depth:
                return "{...}"

            data = copy.deepcopy(value)
            if "openQasm" in data and data["openQasm"] is not None:
                data["openQasm"] = "..."
            return (
                "{"
                + ", ".join(
                    f"{k}: {self._format_value(v, depth + 1, max_depth)}" for k, v in data.items()
                )
                + "}"
            )
        if isinstance(value, list):
            if depth >= max_depth:
                return "[...]"
            return (
                "["
                + ", ".join(self._format_value(item, depth + 1, max_depth) for item in value)
                + "]"
            )
        return repr(value)

    def __repr__(self):
        """Return a string representation of the Result object."""
        out = (
            f"Result(\n"
            f"  device_id={self.device_id},\n"
            f"  job_id={self.job_id},\n"
            f"  success={self.success},\n"
            f"  data={self.data}"
        )
        if self.details:
            for key, value in self.details.items():
                formatted_value = self._format_value(value)
                out += f",\n  {key}={formatted_value}"
        out += "\n)"
        return out



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module providing unified conversions interface between supported
quantum program types.

.. currentmodule:: qbraid.transpiler

Classes
--------

.. autosummary::
   :toctree: ../stubs/

   Conversion
   ConversionGraph
   ConversionScheme

Functions
-----------

.. autosummary::
   :toctree: ../stubs/

   transpile
   translate
   requires_extras

Exceptions
-----------

.. autosummary::
   :toctree: ../stubs/

   ProgramConversionError
   NodeNotFoundError
   ConversionPathNotFoundError

"""
from .annotations import requires_extras
from .converter import translate, transpile
from .edge import Conversion
from .exceptions import ConversionPathNotFoundError, NodeNotFoundError, ProgramConversionError
from .graph import ConversionGraph
from .scheme import ConversionScheme

__all__ = [
    "requires_extras",
    "transpile",
    "translate",
    "Conversion",
    "ConversionGraph",
    "ConversionScheme",
    "ProgramConversionError",
    "NodeNotFoundError",
    "ConversionPathNotFoundError",
]

_lazy_mods = ["conversions"]


def __getattr__(name):
    if name in _lazy_mods:
        import importlib  # pylint: disable=import-outside-toplevel

        module = importlib.import_module(f".{name}", __name__)
        globals()[name] = module
        return module
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


def __dir__():
    return sorted(__all__ + _lazy_mods)
